/***********************************************************************************\
 * Copyright (C) 2004 Steven Deprez                                                *
 *                                                                                 *
 * This file is part of zasm                                                       *
 *                                                                                 *
 * zasm is free software; you can redistribute it and/or modify                    *
 * it under the terms of the GNU General Public License as published by            *
 * the Free Software Foundation; either version 2 of the License, or               *
 * (at your option) any later version.                                             *
 *                                                                                 *
 * This program is distributed in the hope that it will be useful,                 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   *
 * GNU General Public License for more details.                                    *
 *                                                                                 *
 * You should have received a copy of the GNU General Public License               *
 * along with this program; if not, write to the Free Software                     *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.     *
 *                                                                                 *
 * In addition, as a special exception, Peter-Martijn Kuipers gives permission     *
 * to link the code of this program with the appsign library, and distribute       *
 * linked combinations including the two. If you modify this file, you may extend  *
 * this exception to your version of the file, but you are not obligated to do so. *
 * If you do not wish to do so, delete this exception statement form your version. *
\***********************************************************************************/

#define COMMON_C 1
#include <config.h>

#include "converter.h"
#include "../errors/errors.h"

#ifdef HAVE_LIBTIFILES
#include <tilp/tifiles.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef STDC_HEADERS
#include <stdio.h>
#endif

int make_bin(int ct, int vt, char *infile, char *outfile)
{
  struct pages_t *pages;
  struct page_t *page;

  if (!(pages = IH_read_bin(infile)))
    ERROR("internal error\n");
  if (pages->count != 1)
    ERROR("a program should contain exactly one page\n");
  page = pages->pages;
  page->size += 2;
  REALLOCATE(char, page->data, page->size);
  memmove(page->data+2, page->data, page->size - 2);
  page->data[0] = (page->size-2) & 0xff;
  page->data[1] = ((page->size-2)>>8) & 0xff;
  
  if (make_var(ct, vt, page, outfile)<0)
    ERROR("internal error\n");

  free(page->data);
  free(page);
  free(pages);
  return 0;
}

#ifdef HAVE_LIBTIFILES
int make_var(int ct, int vt, struct page_t *page, char *outfile)
{
  Ti8xVarEntry *entry;
  Ti8xRegular *content;
  int e;

  if (!name)
    name = l_basename(outfile);
  //fill entry
  ALLOCATE(Ti8xVarEntry, entry, 1);
  strcpy(entry->name, name);
  entry->type = vt;
  entry->attr = ATTRB_NONE;
  entry->size = page->size;
  entry->data = page->data;
  // fill content
  if (!(content = ti8x_create_regular_content()))
    ERROR("could not create content structure\n");
  content->calc_type = ct;
  strcpy(content->comment, "created with zasm v0.0 alpha");
  content->num_entries = 1;
  content->entries = entry;
  // write out
  if ((e = ti8x_write_regular_file(outfile, content, NULL)) != 0)
    ERROR("couldn't write to file \"%s\": \n\tti8x_write_regular_file (from libtifiles) returned error code %i\n", outfile, e);
  return 0;
}
#else
int make_var(int ct, int vt, struct page_t *page, char *outfile)
{
  FILE *out;

  if (!(out = fopen(outfile, "w")))
    ERROR("couldn't open \"%s\"\n", outfile);
  if (fwrite(page->data, sizeof(char), page->size, out) != page->size)
    ERROR("I/O error\n");
  if (fclose(out))
    ERROR("I/O error\n");
  return 0;
}
#endif
