/***********************************************************************************\
 * Copyright (C) 2004 Steven Deprez                                                *
 *                                                                                 *
 * This file is part of zasm                                                       *
 *                                                                                 *
 * zasm is free software; you can redistribute it and/or modify                    *
 * it under the terms of the GNU General Public License as published by            *
 * the Free Software Foundation; either version 2 of the License, or               *
 * (at your option) any later version.                                             *
 *                                                                                 *
 * This program is distributed in the hope that it will be useful,                 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   *
 * GNU General Public License for more details.                                    *
 *                                                                                 *
 * You should have received a copy of the GNU General Public License               *
 * along with this program; if not, write to the Free Software                     *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.     *
 *                                                                                 *
 * In addition, as a special exception, Peter-Martijn Kuipers gives permission     *
 * to link the code of this program with the appsign library, and distribute       *
 * linked combinations including the two. If you modify this file, you may extend  *
 * this exception to your version of the file, but you are not obligated to do so. *
 * If you do not wish to do so, delete this exception statement form your version. *
\***********************************************************************************/

#define CONVERTER_C 1
#include <config.h>

#include "converter.h"
#include "../errors/errors.h"

#ifdef HAVE_LIBTIFILES
#include <tilp/tifiles.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef STDC_HEADERS
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H 
#include <stdlib.h>
#endif

#include <ctype.h>

#if 0
#define HEX_INT_L(C) ((int)((((C)<='z')&&((C)>='a'))?((C)-'a'+10):0))
#define HEX_INT_U(C) ((int)((((C)<='Z')&&((C)>='A'))?((C)-'A'+10):HEX_INT_L(C)))
#define HEX_INT(C)   ((int)((((C)<='9')&&((C)>='0'))?((C)-'0')   :HEX_INT_U(C)))
#endif

// for compatibility with libappsign
unsigned short int *__ctype_b;
void *__builtin_new(int l)
{
  return malloc(l);
}

// global variables
struct calc_t current_calc = {NULL, 0, NULL};
struct type_t current_type = {NULL, 0, NULL};
char *infile = NULL;
char *outfile = NULL;
char *name;
int typechr = 0;

char *l_basename(char *filename)
{
  int i, j;
  static char base[8];

  i = 0;
  while (((filename[i]<'A') || (filename[i]>'Z')) && ((filename[i]<'a') || (filename[i]>'z')) && filename[i])
    i++;
  if (!filename[i])
    {
      base[0] = 0;
      return base;
    }
  j = 0;
  while ((((filename[i]>='A') && (filename[i]<='Z')) ||
	  ((filename[i]>='a') && (filename[i]<='z')) ||
	  ((filename[i]>='0') && (filename[i]<='9'))) &&
	 j < 7)
    {
      if ((filename[i]>='a') && (filename[i]<='z'))
	base[j] = filename[i] - 'a' + 'A';
      else
	base[j] = filename[i];
      i++; j++;
    }
  base[j] = 0;
  return base;
}

struct type_t types_raw[] =
  {
    {"BIN", 0, make_raw_bin},
    {"HEX", 0, make_raw_hex},

    {NULL, 0, NULL}
  };

#ifdef HAVE_LIBTIFILES
struct type_t types_73[] = 
  {
    {"PRGM", TI73_ASM, make_bin},
    {"PROGRAM", TI73_ASM, make_bin},
    {"PROGRAMME", TI73_ASM, make_bin},
    {"ASM", TI73_ASM, make_bin},
    {"SQSH", TI73_ASM, make_bin},
    {"SQUISH", TI73_ASM, make_bin},
    {"SQUISHED", TI73_ASM, make_bin},
    {"COMPILED", TI73_ASM, make_bin},

    {NULL, 0, NULL}
  };

struct type_t types_82[] = 
  {
    {"PRGM", TI82_PPGM, make_bin},
    {"PROGRAM", TI82_PPGM, make_bin},
    {"PROGRAMME", TI82_PPGM, make_bin},
    {"ASM", TI82_PPGM, make_bin},
    {"SQSH", TI82_PPGM, make_bin},
    {"SQUISH", TI82_PPGM, make_bin},
    {"SQUISHED", TI82_PPGM, make_bin},
    {"COMPILED", TI82_PPGM, make_bin},

    {NULL, 0, NULL}
  };

struct type_t types_83[] = 
  {
    {"PRGM", 0x06, make_bin},       // TI83_ASM is not defined by tifiles
    {"PROGRAM", 0x06, make_bin},
    {"PROGRAMME", 0x06, make_bin},
    {"ASM", 0x06, make_bin},
    {"SQSH", 0x06, make_bin},
    {"SQUISH", 0x06, make_bin},
    {"SQUISHED", 0x06, make_bin},
    {"COMPILED", 0x06, make_bin},

    {NULL, 0, NULL}
  };

struct type_t types_83P[] = 
  {
    {"PRGM", TI83p_PRGM, make_hex83p},
    {"PROGRAM", TI83p_PRGM, make_hex83p},
    {"PROGRAMME", TI83p_PRGM, make_hex83p},

    {"ASM", TI83p_ASM, make_bin83p},
    {"SQSH", TI83p_ASM, make_bin83p},
    {"SQUISH", TI83p_ASM, make_bin83p},
    {"SQUISHED", TI83p_ASM, make_bin83p},
    {"COMPILED", TI83p_ASM, make_bin83p},

    {"APP", 0, make_ti83p_app},
    {"8XK", 0, make_ti83p_8xk},
    {"APPLICATION", 0, make_ti83p_8xk},
    {NULL, 0, NULL}
  };

struct type_t types_85[] = 
  {
    {"PRGM", TI85_STRNG, make_bin},     // binary
    {"PROGRAM", TI85_STRNG, make_bin},
    {"PROGRAMME", TI85_STRNG, make_bin},
    {"ASM", TI85_STRNG, make_bin},
    {"SQSH", TI85_STRNG, make_bin},
    {"SQUISH", TI85_STRNG, make_bin},
    {"SQUISHED", TI85_STRNG, make_bin},
    {"COMPILED", TI85_STRNG, make_bin},
    {"STRNG", TI85_STRNG, make_bin},
    {"STRING", TI85_STRNG, make_bin},

    {NULL, 0, NULL}
  };

struct type_t types_86[] = 
  {
    {"PRGM", TI86_PRGM, make_hex86},      // binary
    {"PROGRAM", TI86_PRGM, make_hex86},
    {"PROGRAMME", TI86_PRGM, make_hex86},
    {"ASM", TI86_PRGM, make_bin86},
    {"SQSH", TI86_PRGM, make_bin86},
    {"SQUISH", TI86_PRGM, make_bin86},
    {"SQUISHED", TI86_PRGM, make_bin86},
    {"COMPILED", TI86_PRGM, make_bin86},

    {NULL, 0, NULL}
  };
#endif
struct calc_t calcs[] = 
  {
    {"RAW", 0, types_raw},
#ifdef HAVE_LIBTIFILES
    {"73", CALC_TI73, types_73},
    {"TI73", CALC_TI73, types_73},

    {"82", CALC_TI82, types_82},
    {"TI82", CALC_TI82, types_82},

    {"83", CALC_TI83, types_83},
    {"TI83", CALC_TI83, types_83},

    {"83PLUS", CALC_TI83P, types_83P},
    {"83P", CALC_TI83P, types_83P},
    {"83+", CALC_TI83P, types_83P},
    {"TI83PLUS", CALC_TI83P, types_83P},
    {"TI83P", CALC_TI83P, types_83P},
    {"TI83+", CALC_TI83P, types_83P},

    {"85", CALC_TI85, types_85},
    {"TI85", CALC_TI85, types_85},

    {"86", CALC_TI86, types_86},
    {"TI86", CALC_TI86, types_86},
#endif
    {NULL, 0, NULL}
  };

struct calc_t find_calc(char *name)
{
  struct calc_t error = {NULL, 0, NULL};
  int i;
  
  for (i = 0; calcs[i].name; i++)
    if (0 == strcasecmp(calcs[i].name, name))
      return calcs[i];
  ERROR("unknown calculator type: \"%s\"\n", name);
  return error;
}

struct type_t find_type(char *name, struct type_t *types)
{
  struct type_t error = {NULL, 0, NULL};
  int i;
  
  if (!types)
    ERROR("no calculator specified yet\n");
  for (i = 0; types[i].name; i++)
    if (0 == strcasecmp(types[i].name, name))
      return types[i];
  ERROR("unknown variable type: \"%s\"\n", name);
  return error;
}

int main(int argc, char *argv[])
{
  unsigned short int **ctype_b = (unsigned short int **)__ctype_b_loc();
  int i;
  __ctype_b = *ctype_b;

  error_init(&argc, &argv);

  for (i = 1; i < argc; i++)
    {
      if ((strcasecmp(argv[i], "--calc") == 0) || (strcasecmp(argv[i], "-c") == 0))
	{
	  i++;
	  if (i >= argc)
	    ERROR("usage (-c|--calc) <calc type> (-t|--type) <variable type> [-b|--begin-chars] -o outfile infile\n");
	  current_calc = find_calc(argv[i]);
	}
      else if ((strcasecmp(argv[i], "--type") == 0) || (strcasecmp(argv[i], "-t") == 0))
	{
	  i++;
	  if (i >= argc)
	    ERROR("usage (-c|--calc) <calc type> (-t|--type) <variable type> [-b|--begin-chars] -o outfile infile\n");
	  current_type = find_type(argv[i], current_calc.types);
	}
      else if ((strcasecmp(argv[i], "--name") == 0) || (strcasecmp(argv[i], "-n") == 0))
	{
	  i++;
	  if (i >= argc)
	    ERROR("usage (-c|--calc) <calc type> (-t|--type) <variable type> [-b|--begin-chars] -o outfile infile\n");
	  name = argv[i];
	}
      else if (strcasecmp(argv[i], "-o") == 0)
	{
	  i++;
	  if (i >= argc)
	    ERROR("usage (-c|--calc) <calc type> (-t|--type) <variable type> [-b|--begin-chars] -o outfile infile\n");
	  outfile = argv[i];
	}
      else if ((strcasecmp(argv[i], "--begin-chars") == 0) || (strcasecmp(argv[i], "-b") == 0))
	typechr = 1;
      else
	infile = argv[i];
    }
  if ((!current_calc.types) || (!current_type.filter) || (!infile) || (!outfile))
    ERROR("not all needed information is given\n");
  if (current_type.filter(current_calc.ID, current_type.ID, infile, outfile))
    ERROR("error while converting \"%s\" to \"%s\"\n", infile, outfile);
  return 0;
}
