/***********************************************************************************\
 * Copyright (C) 2004 Steven Deprez                                                *
 *                                                                                 *
 * This file is part of zasm                                                       *
 *                                                                                 *
 * zasm is free software; you can redistribute it and/or modify                    *
 * it under the terms of the GNU General Public License as published by            *
 * the Free Software Foundation; either version 2 of the License, or               *
 * (at your option) any later version.                                             *
 *                                                                                 *
 * This program is distributed in the hope that it will be useful,                 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   *
 * GNU General Public License for more details.                                    *
 *                                                                                 *
 * You should have received a copy of the GNU General Public License               *
 * along with this program; if not, write to the Free Software                     *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.     *
 *                                                                                 *
 * In addition, as a special exception, Peter-Martijn Kuipers gives permission     *
 * to link the code of this program with the appsign library, and distribute       *
 * linked combinations including the two. If you modify this file, you may extend  *
 * this exception to your version of the file, but you are not obligated to do so. *
 * If you do not wish to do so, delete this exception statement form your version. *
\***********************************************************************************/

#define INTEL_HEX_C 1
#include <config.h>

#include "converter.h"
#include "../errors/errors.h"

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#define HEX_INT_L(C) ((int)((((C)<='z')&&((C)>='a'))?((C)-'a'+10):0))
#define HEX_INT_U(C) ((int)((((C)<='Z')&&((C)>='A'))?((C)-'A'+10):HEX_INT_L(C)))
#define HEX_INT(C)   ((int)((((C)<='9')&&((C)>='0'))?((C)-'0')   :HEX_INT_U(C)))

// returns hex data
struct IH_line_t *IH_read_line(FILE *file)
{
  char line_h[9];
  char line_e[2];
  static struct IH_line_t line[1];
  static char *data = NULL;
  int c;

  //  DEBUG_BEGIN("IH_read_line(%p)\n", file);
  if (fread(line_h, sizeof(char), 9, file) != 9)
    ERROR("I/O error\n");
  if (line_h[0] != ':')
    ERROR("invalid intel hex file\n");
  line->size = (HEX_INT(line_h[1])<<4) + HEX_INT(line_h[2]);
  line->address = (HEX_INT(line_h[3])<<12) + (HEX_INT(line_h[4])<<8) + (HEX_INT(line_h[5])<<4) + HEX_INT(line_h[6]);
  line->type = (HEX_INT(line_h[7])<<4) + HEX_INT(line_h[8]);
  if (line->size)
    {
      REALLOCATE(char, data, line->size * 2);
      if (fread(data, sizeof(char), line->size * 2, file) != line->size * 2)
	ERROR("I/O error\n");
    }
  if (fread(line_e, sizeof(char), 2, file) != 2)
    ERROR("I/O error\n");
  if ((c = fgetc(file)) != EOF)
    {
      if (c == '\r')
	if ((c = fgetc(file)) == EOF)
	  ERROR("I/O error\n");
      if (c != '\n')
	ERROR("invalid intel hex file\n");
    }
  line->data = data;
  line->checksum = (HEX_INT(line_e[0])<<4) + HEX_INT(line_e[1]);
  //DEBUG_END("return %p\n", line);
  return line;
}

struct pages_t *IH_read_bin(char *filename)
{
  struct pages_t *retval;
  struct page_t *pages = NULL;
  int count;
  int size;
  char data[0x10000]; // theoretical max on 1 page: z80 is a 16bit processor
  struct IH_line_t *line;
  FILE *file;
  int i;
  
  if (!(file = fopen(filename, "r")))
    ERROR("couldn't open file \"%s\"\n", filename);
  size = 0;
  count = 0;
  for (line = IH_read_line(file); line; line = IH_read_line(file))
    {
      switch (line->type)
	{
	case 0x00: // normal data line
	  for (i = 0; i < line->size; i++)
	    {
	      if (size+i>=0x10000)
		ERROR("this program contains a page that is too large to be used\n");
	      data[size+i] = (HEX_INT(line->data[i*2])<<4) + HEX_INT(line->data[i*2+1]);
	    }
	  size += line->size;
	  break;
	case 0x01: // EOF
	  goto IH_bin_read;
	case 0x02: // new page;
	  if (size == 0)
	    break; // don't use empty pages
	  count++;
	  REALLOCATE(struct page_t, pages, count);
	  pages[count - 1].size = size;
	  ALLOCATE(char, pages[count-1].data, size);
	  memcpy(pages[count-1].data, data, size);
	  size = 0;
	  break;
	default:
	  ERROR("invalid intel hex file\n");
	}
    }
 IH_bin_read:
  fclose(file);
  if (size != 0)
    {
      count++;
      REALLOCATE(struct page_t, pages, count);
      pages[count - 1].size = size;
      ALLOCATE(char, pages[count-1].data, size);
      memcpy(pages[count-1].data, data, size);
    }
  if (count == 0)
    return NULL;
  ALLOCATE(struct pages_t, retval, 1);
  retval->count = count;
  retval->pages = pages;
  return retval;
}

struct pages_t *IH_read_hex(char *filename)
{
  struct pages_t *retval;
  struct page_t *pages = NULL;
  int count;
  int size;
  char data[0x10000]; // theoretical max on 1 page: z80 is a 16bit processor
  struct IH_line_t *line;
  FILE *file;
  
  if (!(file = fopen(filename, "r")))
    ERROR("couldn't open file \"%s\"\n", filename);
  size = 0;
  count = 0;
  for (line = IH_read_line(file); line; line = IH_read_line(file))
    {
      switch (line->type)
	{
	case 0x00: // normal data line
	  memcpy(data + size, line->data, line->size * 2);
	  size += line->size * 2;
	  break;
	case 0x01: // EOF
	  goto IH_hex_read;
	case 0x02: // new page;
	  if (size == 0)
	    break; // don't use empty pages
	  count++;
	  REALLOCATE(struct page_t, pages, count);
	  pages[count - 1].size = size;
	  ALLOCATE(char, pages[count-1].data, size);
	  memcpy(pages[count-1].data, data, size);
	  size = 0;
	}
    }
 IH_hex_read:
  fclose(file);
  if (size != 0)
    {
      count++;
      REALLOCATE(struct page_t, pages, count);
      pages[count - 1].size = size;
      ALLOCATE(char, pages[count-1].data, size);
      memcpy(pages[count-1].data, data, size);
    }
  if (count == 0)
    return NULL;
  ALLOCATE(struct pages_t, retval, 1);
  retval->count = count;
  retval->pages = pages;
  return retval;
}
