/***********************************************************************************\
 * Copyright (C) 2004 Steven Deprez                                                *
 *                                                                                 *
 * This file is part of zasm                                                       *
 *                                                                                 *
 * zasm is free software; you can redistribute it and/or modify                    *
 * it under the terms of the GNU General Public License as published by            *
 * the Free Software Foundation; either version 2 of the License, or               *
 * (at your option) any later version.                                             *
 *                                                                                 *
 * This program is distributed in the hope that it will be useful,                 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   *
 * GNU General Public License for more details.                                    *
 *                                                                                 *
 * You should have received a copy of the GNU General Public License               *
 * along with this program; if not, write to the Free Software                     *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.     *
 *                                                                                 *
 * In addition, as a special exception, Peter-Martijn Kuipers gives permission     *
 * to link the code of this program with the appsign library, and distribute       *
 * linked combinations including the two. If you modify this file, you may extend  *
 * this exception to your version of the file, but you are not obligated to do so. *
 * If you do not wish to do so, delete this exception statement form your version. *
\***********************************************************************************/

#define TI83P_C 1
#include <config.h>

#include "converter.h"
#include "appsign.h"
#include "../errors/errors.h"

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

int make_ti83p_app(int ct, int vt, char *infile, char *outfile)
{
  int e;
  if ((e = mkApp__FPcN20(infile, outfile, "0104.key")))
    ERROR("mkApp (from libappsign) returned error code %i\n", e);
  return 0;
}

int make_ti83p_8xk(int ct, int vt, char *infile, char *outfile)
{
  int e;
  if ((e = mk8xk__FPcN20(infile, outfile, "0104.key")))
    ERROR("mk8xk (from libappsign) returned error code %i\n", e);
  return 0;
}

int make_bin83p(int ct, int vt, char *infile, char *outfile)
{
  struct pages_t *pages;
  struct page_t *page;

  if (!(pages = IH_read_bin(infile)))
    ERROR("internal error\n");
  if (pages->count != 1)
    ERROR("a program should contain exactly one page\n");
  page = pages->pages;
  if (typechr)
    {
      page->size += 4;
      REALLOCATE(char, page->data, page->size);
      memmove(page->data+4, page->data, page->size - 4);
      page->data[2] = 0xbb; // t2ByteTok
      page->data[3] = 0x6d; // tAsmCmp
    }
  else
    {
      page->size += 2;
      REALLOCATE(char, page->data, page->size);
      memmove(page->data+2, page->data, page->size - 2);
    }
  page->data[0] = (page->size-2) & 0xff;
  page->data[1] = ((page->size-2)>>8) & 0xff;

  if (make_var(ct, vt, page, outfile)<0)
    ERROR("internal error\n");

  free(page->data);
  free(page);
  free(pages);
  return 0;
}

int make_hex83p(int ct, int vt, char *infile, char *outfile)
{
  struct pages_t *pages;
  struct page_t *page;
  char *data;
  int i, size, line_count;

  if (!(pages = IH_read_hex(infile)))
    ERROR("internal error\n");
  if (pages->count != 1)
    ERROR("a program should contain exactly one page\n");
  page = pages->pages;
  
  size = (page->size * 13)/12;
  line_count = page->size / 12;
  ALLOCATE(char, data, size);
  for (i = 0; i < line_count; i++)
    {
      memcpy(data+13*i, page->data+12*i, 12);
      data[13*i + 12] = 0x3f; // newline
    }
  memcpy(data+13*i, page->data+12*i, size - line_count * 13);
  free(page->data);
  free(pages);
  if (typechr)
    {
      size += 5;
      REALLOCATE(char, data, size);
      memmove(data+5, data, size - 5);
      data[2] = 0xbb; // t2ByteTok
      data[3] = 0x6c; // tAsmPrgm
      data[4] = 0x3f; // newline
    }
  else
    {
      size += 2;
      REALLOCATE(char, data, size);
      memmove(data+2, data, size - 2);
    }
  data[0] = (size-2) & 0xff;
  data[1] = ((size-2)>>8) & 0xff;

  page->data = data;
  page->size = size;
  if (make_var(ct, vt, page, outfile)<0)
    ERROR("internal error\n");

  free(page);
  free(data);
  return 0;
}
