#define ZASM_ERRORS_C 1

//#include <config.h>
#include "../config.h"
#ifdef STDC_HEADERS
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include "errors.h"


struct buffer_t
{
  int stack_length;
  int data_length;
  int stack_index;
  int *stack;
  char *data;
};

static struct buffer_t error_buffer = {0,0,0,NULL,NULL};
static struct buffer_t debug_buffer = {0,0,0,NULL,NULL};
static struct buffer_t warning_buffer = {0,0,0,NULL,NULL};

static FILE *error_file = NULL;
static FILE *warning_file = NULL;
static FILE *debug_file = NULL;

char *error_pos_file = NULL;
int error_pos_line_begin = -1;
int error_pos_line_end = -1;

void error_init(int *argc_p, char ***argv_p)
{
  int argc = *argc_p;
  char **argv = *argv_p;
  char **ret_argv;
  int ret_argc = argc;
  int i;

  for (i = 1; i < argc; i++)
    {
      if ((0==strcasecmp("--debug", argv[i])) ||
	  (0==strcasecmp("--warning", argv[i])) ||
	  (0==strcasecmp("--error", argv[i])) ||
	  (0==strcasecmp("--nodebug", argv[i])) ||
	  (0==strcasecmp("--nowarning", argv[i])) ||
	  (0==strcasecmp("--noerror", argv[i])))
	ret_argc--;
      else if ((0==strcasecmp("--debug-file", argv[i])) ||
	  (0==strcasecmp("--warning-file", argv[i])) ||
	  (0==strcasecmp("--error-file", argv[i])))
	{
	  ret_argc-=2;
	  i++;
	}
    }
#define buffer_init(buffer) {buffer.data_length = 256;buffer.stack_length = 64;\
  ALLOCATE(char, buffer.data, buffer.data_length);ALLOCATE(int, buffer.stack, buffer.stack_length);\
  buffer.stack[0] = 0;buffer.stack_index = 0;}

  buffer_init(debug_buffer);
  buffer_init(warning_buffer);
  buffer_init(error_buffer);
#undef buffer_init
  if (ret_argc == argc)
    return;
  ALLOCATE(char *, ret_argv, ret_argc);
  ret_argv[0] = argv[0];
  for (i = 1,ret_argc = 1; i < argc; i++)
    {
      if (0==strcasecmp("--debug", argv[i]))
	debug_file = stderr;
      else if (0==strcasecmp("--warning", argv[i]))
	warning_file = stderr;
      else if (0==strcasecmp("--error", argv[i]))
	error_file = stderr;
      else if (0==strcasecmp("--nodebug", argv[i]))
	debug_file = NULL;
      else if (0==strcasecmp("--nowarning", argv[i]))
	warning_file = NULL;
      else if (0==strcasecmp("--noerror", argv[i]))
	error_file = NULL;
      else if (0==strcasecmp("--debug-file", argv[i]))
	{
	  i++;
	  if (i >= argc)
	    {fprintf(stderr, "error in initialising the error system: missing file name\n");exit(1);}
	  debug_file = fopen(argv[i], "w");
	}
      else if (0==strcasecmp("--warning-file", argv[i]))
	{
	  i++;
	  if (i >= argc)
	    {fprintf(stderr, "error in initialising the error system: missing file name\n");exit(1);}
	  warning_file = fopen(argv[i], "w");
	}
      else if (0==strcasecmp("--error-file", argv[i]))
	{
	  i++;
	  if (i >= argc)
	    {fprintf(stderr, "error in initialising the error system: missing file name\n");exit(1);}
	  error_file = fopen(argv[i], "w");
	}
      else
	{
	  ret_argv[ret_argc] = argv[i];
	  ret_argc++;
	}
    }
  *argc_p = ret_argc;
  *argv_p = ret_argv;
}

#define stack_cur(buffer) buffer->stack[buffer->stack_index]
#define data_cur(buffer) (buffer->data + stack_cur(buffer))

static void error_flush(FILE *file, struct buffer_t *buffer)
{
  if (!file)
    goto written;
  if (fwrite(buffer->data, sizeof(char), stack_cur(buffer), file) != stack_cur(buffer))
    {fprintf(stderr, "can't write messages\n");exit(1);}
 written:
  buffer->stack_index = 0;
  buffer->stack[0] = 0;
  return;
}

static void vbprintf(struct buffer_t *buffer, char *fmt, va_list args)
{
  int e;
  va_list ap;
  if (!buffer->data)
    ERROR("internal error\n");
  if (!buffer->stack)
    ERROR("internal error\n");

  while (1)
    {
      va_copy(ap, args);
      e = vsnprintf(data_cur(buffer), buffer->data_length - stack_cur(buffer), 
		    fmt, ap);
      va_end(ap);
      if (e >= buffer->data_length - stack_cur(buffer))
	{
	  buffer->data_length = e + stack_cur(buffer) + 1;
	  REALLOCATE(char, buffer->data, buffer->data_length);
	}
      else if (e < 0)
	{
	  buffer->data_length += 256;
	  REALLOCATE(char, buffer->data, buffer->data_length);	  
	}
      else
	{
	  stack_cur(buffer)+=e;
	  return;
	}
    }  
}

static void bprintf(struct buffer_t *buffer, char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  vbprintf(buffer, fmt, ap);
  va_end(ap);
}

static void error_push(struct buffer_t *buffer, char *type, char *format, va_list args)
{
  if (type)
    {
      bprintf(buffer, "%s: ", type);
      if (error_pos_file)
	bprintf(buffer, "in \"%s\": ", error_pos_file);
      if (error_pos_line_begin > -1)
	{
	  bprintf(buffer, "%i", error_pos_line_begin);
	  if (error_pos_line_end > -1)
	    bprintf(buffer, " - ");
	  else
	    bprintf(buffer, ": ");
	}
      if (error_pos_line_end > -1)
	bprintf(buffer, "%i: ", error_pos_line_end);
    }
  vbprintf(buffer, format, args);
}

static void error_mark(struct buffer_t *buffer)
{
  buffer->stack_index++;
  if (buffer->stack_index == buffer->stack_length)
    {
      buffer->stack_length += 64;
      REALLOCATE(int, buffer->stack, buffer->stack_length);
    }
  buffer->stack[buffer->stack_index] = buffer->stack[buffer->stack_index - 1];
}

static void error_remove(struct buffer_t *buffer)
{
  if (buffer->stack_index <= 0)
    ERROR("internal error\n");
  buffer->stack_index--;
}

static void error_unmark(struct buffer_t *buffer)
{
  if (buffer->stack_index <= 0)
    ERROR("internal error\n");
  buffer->stack[buffer->stack_index - 1] = buffer->stack[buffer->stack_index];
  buffer->stack_index--;
}

void error_output_debug(char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  error_push(&debug_buffer, NULL, format, ap);
  va_end(ap);
  error_flush(debug_file, &debug_buffer);
}
void error_output_warning(char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  error_push(&warning_buffer, "warning", format, ap);
  va_end(ap);
  error_flush(warning_file, &warning_buffer);
}
void error_output_error(char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  error_push(&error_buffer, "error", format, ap);
  va_end(ap);
  error_flush(error_file, &error_buffer);
  exit(1);
}

void error_push_error(char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  error_push(&error_buffer, "error", format, ap);
  va_end(ap);
}
void error_push_warning(char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  error_push(&warning_buffer, "warning", format, ap);
  va_end(ap);
}
void error_push_debug(char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  error_push(&debug_buffer, NULL, format, ap);
  va_end(ap);
}

void error_mark_error(){error_mark(&error_buffer);}
void error_mark_warning(){error_mark(&warning_buffer);}
void error_mark_debug(){error_mark(&debug_buffer);}

void error_remove_error(){error_remove(&error_buffer);}
void error_remove_warning(){error_remove(&warning_buffer);}
void error_remove_debug(){error_remove(&debug_buffer);}

void error_unmark_error(){error_unmark(&error_buffer);}
void error_unmark_warning(){error_unmark(&warning_buffer);}
void error_unmark_debug(){error_unmark(&debug_buffer);}
