#ifndef ZASM_ERRORS_H
#define ZASM_ERRORS_H 1

//#include <config.h>
#include "../config.h"
#ifdef STDC_HEADERS
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifndef ZASM_ERROR_C
extern char *error_pos_file;
extern int error_pos_line_begin;
extern int error_pos_line_end;
#endif

#ifndef ERROR_NO_SHORT
#define pos_file error_pos_file
#define pos_line_begin error_pos_line_begin
#define pos_line_end error_pos_line_end
// don't use pos_line_begin and/or pos_line end together with pos_line
#define pos_line error_pos_line_end

#define ERROR(format, args...) error_output_error(format, ## args)
#define WARNING(format, args...) error_output_warning(format, ## args)
#define DEBUG(format, args...) error_output_debug(format, ## args)

#define P_ERROR(format, args...) error_push_error(format, ## args)
#define P_WARNING(format, args...) error_push_warning(format, ## args)
#define P_DEBUG(format, args...) error_push_debug(format, ## args)

#define M_ERROR error_mark_error()
#define M_WARNING error_mark_warning()
#define M_DEBUG error_mark_debug()

#define R_ERROR error_remove_error()
#define R_WARNING error_remove_warning()
#define R_DEBUG error_remove_debug()

#define U_ERROR error_unmark_error()
#define U_WARNING error_unmark_warning()
#define U_DEBUG error_unmark_debug()
#endif

#ifndef ERROR_NO_ALLOCATE
#define ALLOCATE(type, var, count) {if (!(var = (type *)malloc(sizeof(type)*(count))))\
 ERROR("couldn't allocate %i instances of %s\n", count, # type);}
#define REALLOCATE(type, var, count) {if (!(var = (type *)realloc(var, sizeof(type)*(count))))\
 ERROR("couldn't allocate %i instances of %s\n", count, # type);}
#endif

void error_init(int *argc_p, char ***argv_p);
void error_output_error(char *format, ...);
void error_output_warning(char *format, ...);
void error_output_debug(char *format, ...);
void error_push_error(char *format, ...);
void error_push_warning(char *format, ...);
void error_push_debug(char *format, ...);
void error_mark_error();
void error_mark_warning();
void error_mark_debug();
void error_remove_error();
void error_remove_warning();
void error_remove_debug();
void error_unmark_error();
void error_unmark_warning();
void error_unmark_debug();

#endif
